@@ -3,6 +3,7 @@
  * make-iso.c: code to generate iso files
  *
  * Copyright (C) 2002-2004 Red Hat, Inc.
+ * Copyright (C) 2005 William Jon McCann <mccann@jhu.edu>
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -20,6 +21,7 @@
  * Boston, MA 02111-1307, USA.
  *
  * Authors: Alexander Larsson <alexl@redhat.com>
+ *          William Jon McCann <mccann@jhu.edu>
  */
 
 #include "config.h"
@@ -40,8 +42,12 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <signal.h>
-#include <libgnomevfs/gnome-vfs.h>
+#include <math.h>
+
+#include <glib.h>
+#include <glib/gstdio.h>
 #include <gtk/gtkmessagedialog.h>
+#include <libgnomevfs/gnome-vfs.h>
 
 #include "nautilus-cd-burner.h"
 #include "make-iso.h"
@@ -83,14 +89,19 @@ make_iso_get_free_space (const char       *filename,
 
 struct mkisofs_output {
 	GMainLoop  *loop;
+	GString    *line;
+	GString    *line_stderr;
+	time_t      start_time;
 	int         result;
 	int         pid;
 	const char *filename;
+	guint64     iso_size;
+	GList      *rates;
 	GError     *error;
 	gboolean    debug;
 };
 
-static struct mkisofs_output *mkisofs_output_ptr;
+struct mkisofs_output *mkisofs_output_ptr;
 
 /**
  * nautilus_burn_make_iso_cancel:
@@ -103,10 +114,8 @@ nautilus_burn_make_iso_cancel (void)
 {
 	if (mkisofs_output_ptr) {
 		kill (mkisofs_output_ptr->pid, SIGINT);
-		unlink (mkisofs_output_ptr->filename);
-		mkisofs_output_ptr->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR, 
-								 NAUTILUS_BURN_ISO_ERROR_GENERAL,
-								 _("The operation was cancelled by the user."));
+		g_unlink (mkisofs_output_ptr->filename);
+		mkisofs_output_ptr->result = NAUTILUS_BURN_RECORDER_RESULT_CANCEL;
 		g_main_loop_quit (mkisofs_output_ptr->loop);
 	}
 }
@@ -136,25 +145,25 @@ copy_file (const char *source,
 {
 	int         sfd, dfd;
 	struct stat stat_buf;
-	char        buffer [1024*8];
+	char        buffer [1024 * 8];
 	ssize_t     len;
 
 	if (link (source, dest) == 0) {
 		return;
 	}
 	
-	if (stat (source, &stat_buf) != 0) {
+	if (g_stat (source, &stat_buf) != 0) {
 		g_warning ("Trying to copy nonexisting file\n");
 		return;
 	}
 
-	sfd = open (source, O_RDONLY);
+	sfd = g_open (source, O_RDONLY, 0);
 	if (sfd == -1) {
 		g_warning ("Can't copy file (open source failed)\n");
 		return;
 	}
 	
-	dfd = open (dest, O_WRONLY | O_CREAT, stat_buf.st_mode);
+	dfd = g_open (dest, O_WRONLY | O_CREAT, stat_buf.st_mode);
 	if (dfd == -1) {
 		close (sfd);
 		g_warning ("Can't copy file (open dest '%s' failed)\n", dest);
@@ -219,11 +228,12 @@ dir_is_empty (const char *virtual_path)
 	g_free (escaped_path);
 	
 	result = gnome_vfs_directory_open (&handle, uri, GNOME_VFS_FILE_INFO_DEFAULT);
+	g_free (uri);
+
 	if (result != GNOME_VFS_OK) {
-		g_free (uri);
 		return TRUE;
 	}
-	
+
 	info = gnome_vfs_file_info_new ();
 
 	found_file = FALSE;
@@ -262,6 +272,7 @@ get_backing_file (const char *virtual_path)
 	escaped_path = gnome_vfs_escape_path_string (virtual_path);
 	uri = g_strconcat ("burn:///", escaped_path, NULL);
 	g_free (escaped_path);
+	/* warning, this can hang on fifos etc */
 	res = gnome_vfs_open (&handle,
 			      uri,
 			      GNOME_VFS_OPEN_READ);
@@ -319,7 +330,7 @@ graft_file_end_dir_visitor (struct mkisofs_state *state)
 			g_free (state->copy_to_dir);
 			state->copy_to_dir = NULL;
 		} else {
-			last_slash = strrchr (state->copy_to_dir, '/');
+			last_slash = strrchr (state->copy_to_dir, G_DIR_SEPARATOR);
 			if (last_slash != NULL) {
 				*last_slash = 0;
 			}
@@ -327,8 +338,27 @@ graft_file_end_dir_visitor (struct mkisofs_state *state)
 	}
 }
 
+/* FIXME: This should probably be an inline */
 static gboolean
-graft_file_visitor (const gchar          *rel_path,
+file_info_is_allowed (GnomeVFSFileInfo *info)
+{
+	g_return_val_if_fail (info != NULL, FALSE);
+
+	/* only allow regular,directory,symlink files */
+	if (info->type != GNOME_VFS_FILE_TYPE_REGULAR
+	    && info->type != GNOME_VFS_FILE_TYPE_DIRECTORY
+	    && info->type != GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK) {
+		return FALSE;
+	} else if (info->type == GNOME_VFS_FILE_TYPE_REGULAR
+		   && !(info->permissions & GNOME_VFS_PERM_ACCESS_READABLE)) {
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static gboolean
+graft_file_visitor (const char           *rel_path,
 		    GnomeVFSFileInfo     *info,
 		    struct mkisofs_state *state,
 		    gboolean             *recurse)
@@ -338,7 +368,12 @@ graft_file_visitor (const gchar          *rel_path,
 	char *copy_path;
 
 	*recurse = TRUE;
-	
+
+	if (! file_info_is_allowed (info)) {
+		g_warning ("Skipping file: %s", rel_path);
+		return TRUE;
+	}
+
 	if (state->copy_to_dir != NULL) {
 		if (info->type == GNOME_VFS_FILE_TYPE_DIRECTORY) {
 			new_copy_dir = g_build_filename (state->copy_to_dir,
@@ -346,7 +381,7 @@ graft_file_visitor (const gchar          *rel_path,
 							 NULL);
 			g_free (state->copy_to_dir);
 			state->copy_to_dir = new_copy_dir;
-			mkdir (state->copy_to_dir, 0777);
+			g_mkdir (state->copy_to_dir, 0777);
 			state->remove_files = g_list_prepend (state->remove_files, g_strdup (state->copy_to_dir));
 			state->copy_depth++;
 		} else {
@@ -405,7 +440,7 @@ graft_file_visitor (const gchar          *rel_path,
 	
 static void
 create_graft_file (GnomeVFSURI          *uri,
-		   const gchar          *prefix,
+		   const char           *prefix,
 		   struct mkisofs_state *state)
 {
 	GnomeVFSFileInfo        *info;
@@ -413,7 +448,10 @@ create_graft_file (GnomeVFSURI          *uri,
 	GnomeVFSResult           result;
 	gboolean                 stop;
 
-	result = gnome_vfs_directory_open_from_uri (&handle, uri, GNOME_VFS_FILE_INFO_DEFAULT);
+	result = gnome_vfs_directory_open_from_uri (&handle,
+						    uri,
+						    GNOME_VFS_FILE_INFO_DEFAULT
+						    | GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS);
 	if (result != GNOME_VFS_OK)
 		return;
 
@@ -421,7 +459,7 @@ create_graft_file (GnomeVFSURI          *uri,
 
 	stop = FALSE;
 	while (! stop) {
-		gchar *rel_path;
+		char    *rel_path;
 		gboolean recurse;
 
 		result = gnome_vfs_directory_read_next (handle, info);
@@ -448,7 +486,7 @@ create_graft_file (GnomeVFSURI          *uri,
 		    && recurse
 		    && info->type == GNOME_VFS_FILE_TYPE_DIRECTORY) {
 			GnomeVFSURI *new_uri;
-			gchar *new_prefix;
+			char        *new_prefix;
 
 			if (prefix == NULL)
 				new_prefix = g_strconcat (info->name, "/",
@@ -515,6 +553,20 @@ mkisofs_stdout_read (GIOChannel   *source,
 	return TRUE;
 }
 
+static void
+mkisofs_output_error (struct mkisofs_output *mkisofs_output,
+		      const char            *message)
+{
+	nautilus_burn_progress_set_fraction (1.0);
+	nautilus_burn_progress_set_time (-1);
+
+	if (!mkisofs_output->error)
+		mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
+							     NAUTILUS_BURN_ISO_ERROR_GENERAL,
+							     message);
+	mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+}
+
 static gboolean  
 mkisofs_stderr_read (GIOChannel   *source,
 		     GIOCondition  condition,
@@ -534,56 +586,50 @@ mkisofs_stderr_read (GIOChannel   *source,
 		}
 
 		if (status == G_IO_STATUS_NORMAL) {
+			char *pos;
+
 			if (strncmp (line, "Total translation table size", 28) == 0) {
 				nautilus_burn_progress_set_fraction (1.0);
 				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_FINISHED;
 			}
 
-			if (strstr (line, "estimate finish")) {
+			if ((pos = strstr (line, "estimate finish"))) {
 				if (sscanf (line, "%6c%% done, estimate finish",
 					    fraction_str) == 1) {
 					fraction_str [6] = 0;
 					fraction = g_strtod (fraction_str, NULL);
-					nautilus_burn_progress_set_fraction (fraction/100.0);
+					nautilus_burn_progress_set_fraction (fraction / 100.0);
+					
+					if (pos) {
+						struct tm tm;
+						time_t    t_finish;
+						long      secs;
+
+						pos += strlen ("estimate finish ");
+						/* mkisofs uses ctime for format time string */
+						strptime (pos, "%a %b %d %H:%M:%S %Y", &tm);
+						t_finish = mktime (&tm);
+						secs = difftime (t_finish, time (NULL));
+						nautilus_burn_progress_set_time (secs);
+					}
 				}
 			}
 
 			if (strstr (line, "Incorrectly encoded string")) {
-				nautilus_burn_progress_set_fraction (1.0);
-				if (!mkisofs_output->error)
-					mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
-										     NAUTILUS_BURN_ISO_ERROR_GENERAL,
-										     _("Incorrectly encoded string."));
-				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+				mkisofs_output_error (mkisofs_output, _("Some files have invalid filenames."));
 			}
 
 			if (strstr (line, "Unknown charset")) {
-				nautilus_burn_progress_set_fraction (1.0);
-				/* FIXME: use better error message after string freeze */
-				if (!mkisofs_output->error)
-					mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
-										     NAUTILUS_BURN_ISO_ERROR_GENERAL,
-										     _("Incorrectly encoded string."));
-				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+				mkisofs_output_error (mkisofs_output, _("Unknown character encoding."));
 			}
 
 			if (strstr (line, "No space left on device")) {
-				nautilus_burn_progress_set_fraction (1.0);
-				if (!mkisofs_output->error)
-					mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
-										     NAUTILUS_BURN_ISO_ERROR_GENERAL,
-										     _("There is no space left on the device."));
-				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+				mkisofs_output_error (mkisofs_output, _("There is no space left on the device."));
 			}
 
 			if (strstr (line, "Value too large for defined data type")) {
 				/* TODO: get filename from error message */
-				nautilus_burn_progress_set_fraction (1.0);
-				if (!mkisofs_output->error)
-					mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
-										     NAUTILUS_BURN_ISO_ERROR_GENERAL,
-										     _("File too large for filesystem."));
-				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+				mkisofs_output_error (mkisofs_output, _("File too large for filesystem."));
 			}
 
 			g_free (line);
@@ -591,6 +637,7 @@ mkisofs_stderr_read (GIOChannel   *source,
 			if (mkisofs_output->debug)
 				g_print ("make_iso stderr: EOF\n");
 			nautilus_burn_progress_set_fraction (1.0);
+			nautilus_burn_progress_set_time (-1);
 			g_main_loop_quit (mkisofs_output->loop);
 			return FALSE;
 		}
@@ -599,6 +646,7 @@ mkisofs_stderr_read (GIOChannel   *source,
 		if (mkisofs_output->debug)
 			g_print ("make_iso stderr: HUP\n");
 		nautilus_burn_progress_set_fraction (1.0);
+		nautilus_burn_progress_set_time (-1);
 		g_main_loop_quit (mkisofs_output->loop);
 		return FALSE;
 	}
@@ -606,6 +654,229 @@ mkisofs_stderr_read (GIOChannel   *source,
 	return TRUE;
 }
 
+static gboolean
+readcd_stderr_read (GIOChannel   *source,
+		    GIOCondition  condition,
+		    gpointer      data)
+{
+	struct mkisofs_output *mkisofs_output = data;
+
+	if (condition & G_IO_IN) {
+		char     *line;
+		char      buf [1];
+		GIOStatus status;
+
+		status = g_io_channel_read_line (source,
+						 &line, NULL, NULL, NULL);
+
+		if (line && mkisofs_output->debug) {
+			g_print ("readcd stderr: %s", line);
+		}
+
+		if (status == G_IO_STATUS_NORMAL) {
+			char *pos;
+
+			if (mkisofs_output->line_stderr) {
+				g_string_append (mkisofs_output->line_stderr, line);
+				g_free (line);
+				line = g_string_free (mkisofs_output->line_stderr, FALSE);
+				mkisofs_output->line_stderr = NULL;
+			}
+
+			pos = strstr (line, "addr:");
+			if (pos) {
+				guint64 byte;
+				double  fraction;
+
+				pos += strlen ("addr:");
+				byte = (guint64) atol (pos) * 2048; /* reports blocks of 2048 bytes */
+				if (mkisofs_output->iso_size > 0) {
+					fraction = (double) byte / mkisofs_output->iso_size;
+					nautilus_burn_progress_set_fraction (fraction);
+					nautilus_burn_progress_set_time (-1);
+				}
+			}
+
+			if (strstr (line, "Time total:")) {
+				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_FINISHED;
+			}
+
+			g_free (line);
+		} else if (status == G_IO_STATUS_AGAIN) {
+			/* A non-terminated line was read, read the data into the buffer. */
+			status = g_io_channel_read_chars (source, buf, 1, NULL, NULL);
+			if (status == G_IO_STATUS_NORMAL) {
+				if (mkisofs_output->line_stderr == NULL) {
+					mkisofs_output->line_stderr = g_string_new (NULL);
+				}
+				g_string_append_c (mkisofs_output->line_stderr, buf [0]);
+			}
+		} else if (status == G_IO_STATUS_EOF) {
+			g_main_loop_quit (mkisofs_output->loop);
+
+			return FALSE;
+		}
+	} else if (condition & G_IO_HUP) {
+		/* only handle the HUP when we have read all available lines of output */
+		if (mkisofs_output->debug)
+			g_print ("readcd stderr: HUP\n");
+		g_main_loop_quit (mkisofs_output->loop);
+
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static gdouble
+get_average_rate (GList  **rates,
+		  gdouble  rate)
+{
+	const unsigned int max_num = 16;
+	const unsigned int scale  = 1000;
+	gdouble            average;
+	gint32             int_rate;
+	GList             *l;
+
+	if (g_list_length (*rates) > max_num) {
+		*rates = g_list_delete_link (*rates, *rates);
+	}
+	int_rate = (gint32)round (scale * rate);
+
+	*rates = g_list_append (*rates, GINT_TO_POINTER (int_rate));
+
+	average = 0;
+	for (l = *rates; l != NULL; l = l->next) {
+		gdouble r = (gdouble)GPOINTER_TO_INT (l->data) / scale;
+		average += r;
+	}
+
+	average /= g_list_length (*rates);
+
+	return average;
+}
+
+static gboolean
+cdrdao_stderr_line (struct mkisofs_output *mkisofs_output,
+		    const char            *line)
+{
+	int t1, t2, s1, s2, s3;
+	int min, sec, sub;
+
+	if (line && mkisofs_output->debug) {
+		g_print ("cdrdao stderr: %s", line);
+	}
+
+	if (sscanf (line, "Copying audio tracks %d-%d: start %d:%d:%d, length %d:%d:%d",
+		    &t1, &t2, &s1, &s2, &s3, &min, &sec, &sub) == 8) {
+		mkisofs_output->iso_size = (guint64) min * 60 + sec;
+	}
+
+	if (sscanf (line, "%d:%d:%d", &min, &sec, &sub) == 3) {
+		if (mkisofs_output->iso_size > 0) {
+			gdouble fraction;
+			guint64 secs = (guint64) min * 60 + sec;
+
+			if (mkisofs_output->start_time > 0) {
+				guint64 elapsed;
+				guint64 remaining;
+				gdouble rate;
+
+				elapsed = time (NULL) - mkisofs_output->start_time;
+				rate = (gdouble)secs / (gdouble)elapsed;
+
+				if (rate > 0) {
+					gdouble ave_rate;
+
+					ave_rate = get_average_rate (&mkisofs_output->rates, rate);
+					remaining = (mkisofs_output->iso_size - secs) / ave_rate;
+					nautilus_burn_progress_set_time (remaining);
+				}
+
+			}
+
+			fraction = (gdouble) secs / mkisofs_output->iso_size;
+			nautilus_burn_progress_set_fraction (fraction);
+		}
+	}
+
+	if (strstr (line, "toc and track data finished successfully")) {
+		mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_FINISHED;
+	}
+
+	return TRUE;
+}
+
+static gboolean
+cdrdao_stderr_read (GIOChannel   *source,
+		    GIOCondition  condition,
+		    gpointer      data)
+{
+	struct mkisofs_output *mkisofs_output = data;
+	gboolean               res            = TRUE;
+
+	if (condition & G_IO_IN) {
+		char     *line;
+		char      buf [1];
+		GIOStatus status;
+
+		status = g_io_channel_read_line (source,
+						 &line, NULL, NULL, NULL);
+
+		if (status == G_IO_STATUS_NORMAL) {
+			if (mkisofs_output->line_stderr) {
+				g_string_append (mkisofs_output->line_stderr, line);
+				g_free (line);
+				line = g_string_free (mkisofs_output->line_stderr, FALSE);
+				mkisofs_output->line_stderr = NULL;
+			}
+
+			res = cdrdao_stderr_line (mkisofs_output, line);
+			
+			g_free (line);
+		} else if (status == G_IO_STATUS_AGAIN) {
+			/* A non-terminated line was read, read the data into the buffer. */
+			status = g_io_channel_read_chars (source, buf, 1, NULL, NULL);
+
+			if (status == G_IO_STATUS_NORMAL) {
+				char *line2;
+
+				if (mkisofs_output->line_stderr == NULL) {
+					mkisofs_output->line_stderr = g_string_new (NULL);
+				}
+				g_string_append_c (mkisofs_output->line_stderr, buf [0]);
+
+				switch (buf [0]) {
+				case '\n':
+				case '\r':
+				case '\xe2':
+				case '\0':
+					line2 = g_string_free (mkisofs_output->line_stderr, FALSE);
+					mkisofs_output->line_stderr = NULL;
+					res = cdrdao_stderr_line (mkisofs_output, line2);
+					g_free (line2);
+					break;
+				default:
+					break;
+				}
+			}
+		} else if (status == G_IO_STATUS_EOF) {
+			g_main_loop_quit (mkisofs_output->loop);
+
+			res = FALSE;
+		}
+	} else if (condition & G_IO_HUP) {
+		/* only handle the HUP when we have read all available lines of output */
+		if (mkisofs_output->debug)
+			g_print ("cdrdao stderr: HUP\n");
+		g_main_loop_quit (mkisofs_output->loop);
+
+		res = FALSE;
+	}
+
+	return res;
+}
+
 /* this is an ugly hack until utf8/iconv support is added into upstream mkisofs */
 static gboolean
 ncb_mkisofs_supports_utf8 (void)
@@ -625,25 +896,191 @@ ncb_mkisofs_supports_utf8 (void)
 	return supported;
 }
 
+static gboolean
+start_async_with_watch (char        **args,
+			GPid         *ppid,
+			GIOFunc       out_watch_func,
+			GIOFunc       err_watch_func,
+			gpointer      user_data,
+			guint        *out_watch_id,
+			guint        *err_watch_id,
+			int          *input_pipe,
+			GError      **error)
+{
+	gboolean    ret;
+	int	    stdin_pipe;
+	int	    stdout_pipe;
+	int	    stderr_pipe;
+	GPid	    pid = 0;
+
+	g_return_val_if_fail (args != NULL, FALSE);
+
+	ret = g_spawn_async_with_pipes (NULL,
+					args,
+					NULL,
+					G_SPAWN_SEARCH_PATH,
+					NULL,
+					NULL,
+					&pid,
+					&stdin_pipe,
+					&stdout_pipe,
+					&stderr_pipe,
+					error);
+
+	if (!ret)
+		return FALSE;
+
+	if (ppid)
+		*ppid = pid;
+
+	if (input_pipe)
+		*input_pipe = stdin_pipe;
+
+	fcntl (stdout_pipe, F_SETFL, O_NONBLOCK);
+	fcntl (stderr_pipe, F_SETFL, O_NONBLOCK);
+
+	if (out_watch_func) {
+		GIOChannel *channel;
+		guint	    id;
+
+		channel = g_io_channel_unix_new (stdout_pipe);
+		g_io_channel_set_flags (channel,
+					g_io_channel_get_flags (channel) | G_IO_FLAG_NONBLOCK,
+					NULL);
+		g_io_channel_set_encoding (channel, NULL, NULL);
+
+		id = g_io_add_watch (channel,
+				     G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+				     out_watch_func,
+				     user_data);
+
+		g_io_channel_unref (channel);
+
+		if (out_watch_id)
+			*out_watch_id = id;
+	} else {
+		if (out_watch_id)
+			*out_watch_id = -1;
+	}
+
+	if (err_watch_func) {
+		GIOChannel *channel;
+		guint	    id;
+
+		channel = g_io_channel_unix_new (stderr_pipe);
+		g_io_channel_set_flags (channel,
+					g_io_channel_get_flags (channel) | G_IO_FLAG_NONBLOCK,
+					NULL);
+		g_io_channel_set_encoding (channel, NULL, NULL);
+
+		id = g_io_add_watch (channel,
+				     G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+				     err_watch_func,
+				     user_data);
+
+		g_io_channel_unref (channel);
+
+		if (err_watch_id)
+			*err_watch_id = id;
+	} else {
+		if (err_watch_id)
+			*err_watch_id = -1;
+	}
+
+	return ret;
+}
+
+static void
+run_process (struct mkisofs_output    *mkisofs_output,
+	     GIOFunc                   out_watch_func,
+	     GIOFunc                   err_watch_func,
+	     GPtrArray                *argv)
+{
+	int         res;
+	guint       stdout_tag, stderr_tag;
+	GError     *error;
+
+	mkisofs_output->line_stderr = NULL;
+
+ retry:
+	mkisofs_output->result     = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+	mkisofs_output->line       = NULL;
+	mkisofs_output->start_time = time (NULL);
+
+	if (mkisofs_output->line_stderr != NULL) {
+		g_string_truncate (mkisofs_output->line_stderr, 0);
+	} else {
+		mkisofs_output->line_stderr = g_string_new (NULL);
+	}
+
+	nautilus_burn_progress_set_text (_("Creating disc image"));
+	nautilus_burn_progress_set_fraction (0.0);
+	nautilus_burn_progress_set_image_spinning (TRUE);
+
+	if (mkisofs_output->debug) {
+		guint i;
+		g_print ("launching command: ");
+		for (i = 0; i < argv->len - 1; i++) {
+			g_print ("%s ", (char *) g_ptr_array_index (argv, i));
+		}
+		g_print ("\n");
+	}
+
+	error = NULL;
+
+	res = start_async_with_watch ((char **)argv->pdata,
+				      &mkisofs_output->pid,
+				      out_watch_func,
+				      err_watch_func,
+				      mkisofs_output,
+				      &stdout_tag,
+				      &stderr_tag,
+				      NULL,
+				      &error);
+
+	if (! res) {
+		g_warning ("command failed: %s\n", error->message);
+		g_error_free (error);
+		/* TODO: Better error handling */
+	} else {
+		mkisofs_output->loop = g_main_loop_new (NULL, FALSE);
+
+		g_main_loop_run (mkisofs_output->loop);
+		g_main_loop_unref (mkisofs_output->loop);
+		
+		if (stdout_tag > 0)
+			g_source_remove (stdout_tag);
+		if (stderr_tag > 0)
+			g_source_remove (stderr_tag);
+
+		if (mkisofs_output->result == NAUTILUS_BURN_RECORDER_RESULT_RETRY) {
+			goto retry;
+		}
+	}
+
+	g_list_free (mkisofs_output->rates);
+
+	nautilus_burn_progress_set_image_spinning (FALSE);
+}
+
 /**
  * nautilus_burn_make_iso:
- * @filename:
- * @label:
- * @warn_low_space:
- * @use_joliet:
- * @debug:
+ * @filename: name of a file to use for the image
+ * @label: a string to use as the label for the image
+ * @flags: #NautilusBurnImageCreateFlags
+ * @type: #NautilusBurnImageType
+ * @error: a return location for errors
  *
  * Create an ISO image in filename from the data files in burn:///
  *
- * Return value: 
+ * Return value: #NautilusBurnRecorderResult
  **/
 int
-nautilus_burn_make_iso (const char *filename,
-			const char *label,
-			gboolean    warn_low_space,
-			gboolean    use_joliet,
-			gboolean    debug,
-			GError    **error)
+nautilus_burn_make_iso (const char                  *filename,
+			const char                  *label,
+			NautilusBurnImageCreateFlags flags,
+			NautilusBurnImageType       *type,
+			GError                     **error)
 {
 	GnomeVFSURI          *uri;
 	char                 *filelist = NULL;
@@ -663,6 +1100,14 @@ nautilus_burn_make_iso (const char *filename,
 	guint64               iso_size;
 	GnomeVFSFileSize      size;
 	gboolean              has_utf8_support;
+	gboolean              use_joliet;
+
+	use_joliet = (flags & NAUTILUS_BURN_IMAGE_CREATE_JOLIET);
+
+	if (type)
+		*type = NAUTILUS_BURN_IMAGE_TYPE_ISO9660;
+
+	memset (&mkisofs_output, 0, sizeof (mkisofs_output));
 
 	if (label && (strlen (label) > 32)) {
 		g_set_error (error,
@@ -673,7 +1118,7 @@ nautilus_burn_make_iso (const char *filename,
 	}
 
 	mkisofs_output.error = NULL;
-	mkisofs_output.debug = debug;
+	mkisofs_output.debug = flags & NAUTILUS_BURN_IMAGE_CREATE_DEBUG;
 
 	dirname = g_strdup_printf ("iso-%s.XXXXXX", g_get_user_name ());
 	tempdir = g_build_filename (g_get_tmp_dir (), dirname, NULL);
@@ -691,7 +1136,7 @@ nautilus_burn_make_iso (const char *filename,
 	}
 	
 	state.emptydir = g_build_filename (state.tmpdir, "emptydir", NULL);
-	mkdir (state.emptydir, 0777);
+	g_mkdir (state.emptydir, 0777);
 	state.remove_files = g_list_prepend (state.remove_files, g_strdup (state.emptydir));
 
 	filelist = g_build_filename (state.tmpdir, "filelist", NULL);
@@ -743,7 +1188,7 @@ nautilus_burn_make_iso (const char *filename,
 
 	sub_error = NULL;
 
-	if (debug) {
+	if (flags & NAUTILUS_BURN_IMAGE_CREATE_DEBUG) {
 		g_print ("launching command: ");
 		for (i = 0; argv [i] != NULL; i++) {
 			g_print ("%s ", argv [i]);
@@ -795,10 +1240,8 @@ nautilus_burn_make_iso (const char *filename,
 
 	if (res == -1) {
 		g_warning ("Cannot get free space at %s\n", dirname);
-		g_free (dirname);
 	} else if (iso_size > size) {
-		g_free (dirname);
-		if (warn_low_space) {
+		if (flags & NAUTILUS_BURN_IMAGE_CREATE_WARN_SPACE) {
 			mkisofs_output.error = g_error_new (NAUTILUS_BURN_ISO_ERROR, 
 							    NAUTILUS_BURN_ISO_ERROR_GENERAL,
 							    _("The selected location does not have enough space to store the disc image (%ld MiB needed)."),
@@ -808,9 +1251,13 @@ nautilus_burn_make_iso (const char *filename,
 			mkisofs_output.result = NAUTILUS_BURN_RECORDER_RESULT_RETRY;
 		}
 
+		g_free (dirname);
+
 		goto cleanup;
 	}
 
+	g_free (dirname);
+
 	i = 0;
 	argv [i++] = "mkisofs";
 	argv [i++] = "-r";
@@ -832,7 +1279,7 @@ nautilus_burn_make_iso (const char *filename,
 	argv [i++] = filename;
 	argv [i++] = NULL;
 
-	if (debug) {
+	if (flags & NAUTILUS_BURN_IMAGE_CREATE_DEBUG) {
 		g_print ("launching command: ");
 		for (i = 0; argv [i] != NULL; i++) {
 			g_print ("%s ", argv [i]);
@@ -840,7 +1287,7 @@ nautilus_burn_make_iso (const char *filename,
 		g_print ("\n");
 	}
 
-	nautilus_burn_progress_set_text (_("Creating CD image"));
+	nautilus_burn_progress_set_text (_("Creating disc image"));
 	nautilus_burn_progress_set_fraction (0.0);
 	nautilus_burn_progress_set_image_spinning (TRUE);
 	sub_error = NULL;
@@ -892,17 +1339,213 @@ nautilus_burn_make_iso (const char *filename,
 
  cleanup:
 	for (l = state.remove_files; l != NULL; l = l->next) {
-		remove ((char *)l->data);
+		g_remove ((char *)l->data);
 		g_free (l->data);
 	}
 	g_list_free (state.remove_files);
 	
 	g_free (filelist);
 	g_free (state.emptydir);
-	rmdir (tempdir);
+	g_rmdir (tempdir);
 	g_free (tempdir);
 	nautilus_burn_progress_set_image_spinning (FALSE);
 
+	if (mkisofs_output.result == NAUTILUS_BURN_RECORDER_RESULT_ERROR) {
+		if (!mkisofs_output.error)
+			mkisofs_output.error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
+								    NAUTILUS_BURN_ISO_ERROR_GENERAL,
+								    _("Unknown error"));
+		g_unlink (mkisofs_output.filename);
+
+		g_propagate_error (error, mkisofs_output.error);
+	}
+
+	return mkisofs_output.result;
+}
+
+/**
+ * nautilus_burn_make_iso_from_drive:
+ * @filename: name of a file to use for the image
+ * @drive: #NautilusBurnDrive from which to read the source media
+ * @warn_low_space: set %TRUE issue a warning when disk space is low
+ *
+ * Create an ISO image in filename from the data in @drive
+ *
+ * Return value: #NautilusBurnRecorderResult
+ **/
+int
+nautilus_burn_make_iso_from_drive (const char                  *filename,
+				   const NautilusBurnDrive     *drive,
+				   NautilusBurnImageCreateFlags flags,
+				   NautilusBurnImageType       *type,
+				   char                       **toc_filename,
+				   GError                     **error)
+{
+	GPtrArray            *argv = NULL;
+	GError               *sub_error;
+	struct mkisofs_output mkisofs_output;
+	char                 *device_arg, *outfile_arg;
+	char                 *tocfile_arg    = NULL;
+	char                 *stdout_data, *stderr_data;
+	char                 *pos;
+	char                 *dirname;
+	int                   exit_status, res;
+	guint64               iso_size;
+	GnomeVFSFileSize      size;
+	NautilusBurnMediaType media_type;
+	gboolean              is_rewritable, is_blank, has_data, has_audio;
+	GIOFunc               out_watch_func = NULL;
+	GIOFunc               err_watch_func = NULL;
+
+	memset (&mkisofs_output, 0, sizeof (mkisofs_output));
+
+	if (toc_filename)
+		*toc_filename = NULL;
+	if (type)
+		*type = NAUTILUS_BURN_IMAGE_TYPE_UNKNOWN;
+
+	mkisofs_output.error    = NULL;
+	mkisofs_output.debug    = (flags & NAUTILUS_BURN_IMAGE_CREATE_DEBUG);
+	mkisofs_output.result   = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+	mkisofs_output.filename = filename;
+
+	media_type = nautilus_burn_drive_get_media_type_full (drive,
+							      &is_rewritable,
+							      &is_blank,
+							      &has_data,
+							      &has_audio);
+	if (has_audio) {
+		device_arg = g_strdup_printf ("%s", drive->cdrecord_id);
+		outfile_arg = g_strdup_printf ("%s", filename);
+		tocfile_arg = g_strdup_printf ("%s.toc", filename);
+		if (toc_filename)
+			*toc_filename = g_strdup (tocfile_arg);
+		if (type)
+			*type = NAUTILUS_BURN_IMAGE_TYPE_BINCUE;
+
+		argv = g_ptr_array_new ();
+		g_ptr_array_add (argv, "cdrdao");
+		g_ptr_array_add (argv, "disk-info");
+		g_ptr_array_add (argv, "--device");
+		g_ptr_array_add (argv, device_arg);
+		g_ptr_array_add (argv, NULL);
+		
+	} else {
+		device_arg = g_strdup_printf ("-dev=%s", drive->cdrecord_id);
+		outfile_arg = g_strdup_printf ("-f=%s", filename);
+		if (toc_filename)
+			*toc_filename = g_strdup_printf ("%s.toc", filename);
+		if (type)
+			*type = NAUTILUS_BURN_IMAGE_TYPE_ISO9660;
+
+		argv = g_ptr_array_new ();
+		g_ptr_array_add (argv, "readcd");
+		g_ptr_array_add (argv, "-sectors=0-0");
+		g_ptr_array_add (argv, device_arg);
+		g_ptr_array_add (argv, outfile_arg);
+		g_ptr_array_add (argv, NULL);
+	}
+
+	sub_error = NULL;
+
+	if (!g_spawn_sync (NULL,
+			   (char **)argv->pdata,
+			   NULL,
+			   G_SPAWN_SEARCH_PATH,
+			   NULL, NULL,
+			   &stdout_data,
+			   &stderr_data,
+			   &exit_status,
+			   &sub_error)) {
+		mkisofs_output.error = g_error_new (NAUTILUS_BURN_ISO_ERROR, 
+						    NAUTILUS_BURN_ISO_ERROR_GENERAL,
+						    _("Could not run sub process: %s."),
+						    sub_error->message);
+		g_error_free (sub_error);
+		mkisofs_output.result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+		goto cleanup;
+	}
+
+	g_ptr_array_free (argv, TRUE);
+	g_free (stdout_data);
+
+	if (has_audio) {
+		/* assume audio CD is 800 MiB */
+		iso_size = 838860800;
+	} else {
+		pos = strstr (stderr_data, "Capacity:");
+		if (pos) {
+			pos += strlen ("Capacity:");
+			iso_size = (guint64) atol (pos) * 2048; /* reports blocks of 2048 bytes */
+		} else {
+			iso_size = 0;
+		}
+		mkisofs_output.iso_size = iso_size;
+	}
+
+	g_free (stderr_data);
+
+	dirname = g_path_get_dirname (filename);
+	res = make_iso_get_free_space (dirname, &size);
+
+	if (res == -1) {
+		g_warning ("Cannot get free space at %s\n", dirname);
+		g_free (dirname);
+	} else if (iso_size > size) {
+		g_free (dirname);
+		if (flags & NAUTILUS_BURN_IMAGE_CREATE_WARN_SPACE) {
+			mkisofs_output.error = g_error_new (NAUTILUS_BURN_ISO_ERROR, 
+							    NAUTILUS_BURN_ISO_ERROR_GENERAL,
+							    _("The selected location does not have enough space to store the disc image (%ld MiB needed)."),
+							    (unsigned long)iso_size / 1048576);
+			mkisofs_output.result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+		} else {
+			mkisofs_output.result = NAUTILUS_BURN_RECORDER_RESULT_RETRY;
+		}
+
+		goto cleanup;
+	}
+
+	if (has_audio) {
+		argv = g_ptr_array_new ();
+		g_ptr_array_add (argv, "cdrdao");
+		g_ptr_array_add (argv, "read-cd");
+		g_ptr_array_add (argv, "--read-raw");
+		g_ptr_array_add (argv, "--datafile");
+		g_ptr_array_add (argv, outfile_arg);
+		g_ptr_array_add (argv, "--device");
+		g_ptr_array_add (argv, device_arg);
+		g_ptr_array_add (argv, "-v");
+		g_ptr_array_add (argv, "2");
+		g_ptr_array_add (argv, tocfile_arg);
+		g_ptr_array_add (argv, NULL);
+
+		out_watch_func = NULL;
+		err_watch_func = cdrdao_stderr_read;
+	} else {
+		argv = g_ptr_array_new ();
+		g_ptr_array_add (argv, "readcd");
+		g_ptr_array_add (argv, device_arg);
+		g_ptr_array_add (argv, outfile_arg);
+		g_ptr_array_add (argv, NULL);
+
+		out_watch_func = NULL;
+		err_watch_func = readcd_stderr_read;
+	}
+
+	mkisofs_output_ptr = &mkisofs_output;
+	run_process (&mkisofs_output,
+		     out_watch_func,
+		     err_watch_func,
+		     argv);
+	mkisofs_output_ptr = NULL;
+
+ cleanup:
+	g_free (outfile_arg);
+	g_free (tocfile_arg);
+	g_free (device_arg);
+	g_ptr_array_free (argv, TRUE);
+
 	if (mkisofs_output.result == NAUTILUS_BURN_RECORDER_RESULT_ERROR) {
 		if (!mkisofs_output.error)
 			mkisofs_output.error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
@@ -956,9 +1599,9 @@ ncb_make_valid_utf8 (const char *name)
 
 /**
  * nautilus_burn_verify_iso:
- * @filename:
- * @iso_label:
- * @error:
+ * @filename: name of a file to use for the image
+ * @iso_label: return location for the image label
+ * @error: return location for errors
  *
  * Verify that filename is a valid ISO image
  *
@@ -1036,7 +1679,7 @@ nautilus_burn_verify_iso (const char *filename,
  bail:
 	*error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
 				      NAUTILUS_BURN_ISO_ERROR_GENERAL,
-				      _("Not a valid CD image."));
+				      _("Not a valid disc image."));
 
 	return FALSE;
 }
